(function (factory) {
    if (typeof define === 'function' && define.amd) {
        // AMD. Register as an anonymous module.
        define(['jquery'], factory);
    } else if (typeof module === 'object' && module.exports) {
        // Node/CommonJS
        module.exports = factory(require('jquery'));
    } else {
        // Browser globals
        window.waterfall = factory(jQuery);
    }
}(function ($) {
   /**
    * --------------------------------------------
    * 可以在初始化时设置的参数
    * id ：组件id
    * marginOffset ：每个item之间的间隔
    * itemContent ：选项内容生成方法
    * 且itemContent方法的返回值必须遵循以下规则
    * {
    *     html:'',//需要显示在item中的html代码
    *     types:['image','image'],//item中涉及资源的类型名称数组，暂时只支持图片资源（待优化）
    *     resources:['','']//item中涉及资源的路径数组
    * }
    * itemWidth ：item宽度
    * --------------------------------------------
    * columnHeight ：瀑布流各行的宽度累积值数组
    * columnNum ：瀑布流行数
    * index ：循环索引，用于协调全部刷新和局部刷新，局部刷新和局部刷新
    * boxWidth ：itemWidth+marginOffset
    * $boxes ：所有子元素
    * $con ：瀑布流布局容器
    * --------------------------------------------
    * 可能会出现定位混乱的情况，尽量不要同时执行addSome或者addOne（待优化）
    * addOne ：新增一个item
    * addSome ：新增多个item
    * empty ：清空，再不改变定位参数的情况下清空数据
    * --------------------------------------------
    * 响应式瀑布流布局工具
    */
    var waterfall = function (options) {

        var component = {};//组件
        var inner = $.extend({},options);//在代码里边用inner代替options，防止暴露过多的属性

        //参数校验
        inner.error = [];
        if(!inner.id){
            inner.error.push('组件id不能为空');
        }
        if(!inner.itemContent){
            inner.error.push('选项内容生成方法itemContent不能为空');
        }
        if(typeof inner.itemContent!=='function'){
            inner.error.push('选项内容生成方法itemContent类型必须是function');
        }
        if(inner.error.length>0){
            console.log('jquery plugin waterfall init error:'+inner.error);
            return false;
        }

        //参数默认值
        if(!inner.marginOffset){
            inner.marginOffset = 10;
        }
        if(!inner.itemWidth){
            inner.itemWidth = 200;
        }

        /* 清空，再不改变定位参数的情况下清空数据 */
        component.empty = function(){
            inner.$con.empty();
            //属性重置
            inner.columnHeight = [];
            inner.index = 0;
            inner.error = [];
        };
        /* 新增一个item */
        component.addOne = function(itemData){
            var item = inner.itemContent(itemData);
            inner.$con.append('<div>'+item.html+'</div>');
            var items = [];
            items.push(item);
            loadResource(items);
        };
        /* 新增多个item */
        component.addSome = function(data){
            if(data&&data.length){
                var items = [];
                for(var i=0;i<data.length;i++){
                    items.push(inner.itemContent(data[i]));
                }
                var html = '';
                for(var i=0;i<items.length;i++){
                    html += '<div>'+items[i].html+'</div>';
                }
                inner.$con.append(html);
                loadResource(items);
            }
        };

        /* 加载资源 */
        function loadResource(items){
            var count = 0;
            for(var i=0;i<items.length;i++){
                count += items[i].types.length;
            }
            var result = 0;
            if(count===result){
                refreshNewItem();//如果没有任何需要加载的资源则立即执行刷新定位
            }
            for(var i=0;i<items.length;i++){
                var item = items[i];
                var types = item.types;
                for(var j=0;j<types.length;j++){
                    var type = types[j];
                    //处理图片资源
                    if(type==='image'){
                        var img = new Image();
                        img.onload = function(){
                            result++;
                            if(count===result){
                                refreshNewItem();//如果资源全部加载完毕，则刷新定位新增item
                            }
                        }
                        img.onerror=function(){
                            result++;
                            if(count===result){
                                refreshNewItem();//如果资源全部加载完毕，则刷新定位新增item
                            }
                        }
                        img.src = item.resources[j];
                    }
                }
            }
        }

        /* 刷新新增item */
        function refreshNewItem(){
            var newItems = inner.$con.children('div:not([style])');
            newItems.css({
                'position':'absolute',
                'width':inner.itemWidth+'px'
            });
            //防止还未初始化
            if(!inner.columnHeight){
                inner.index = 0;
                inner.columnHeight=[];
                var bodyWidth = document.body.offsetWidth;
                var n = parseInt(bodyWidth/inner.boxWidth);
                inner.columnNum = n==0?1:n;
            }
            var i = 0;
            var l = newItems.length;
            for (i = 0; i <l ; i++) {
                if ((i+inner.index)<inner.columnNum) {
                    inner.columnHeight[i+inner.index]=newItems[i].offsetHeight+inner.marginOffset;
                    newItems.eq(i).css({
                        'top':0,
                        'left':(i+inner.index)*inner.boxWidth+inner.marginOffset+'px'
                    });
                } else{
                    var innsertColumn = min(inner.columnHeight);
                    var imgHeight = newItems[i].offsetHeight+inner.marginOffset;
                    newItems.eq(i).css({
                        'top':inner.columnHeight[innsertColumn]+'px',
                        'left':innsertColumn*inner.boxWidth+inner.marginOffset+'px'
                    });
                    inner.columnHeight[innsertColumn] +=imgHeight;
                };
            };
            /* 重新定位之后需要重新计算父容器的高度，如果不明确赋值高度，会出现父容器高度为0，然后页面中的某些元素遮住瀑布流容器的情况 */
            if(inner.columnHeight.length>0){
                inner.$con.css('height',inner.columnHeight[max(inner.columnHeight)]+'px');
            }
            inner.index += i;
        };

        /* 刷新全部item */
        function refreshAll() {
            //获取所有item
            inner.$boxes = inner.$con.children('div');
            inner.$boxes.css({
                'position':'absolute',
                'width':inner.itemWidth+'px'
            });
            inner.index = 0;
            inner.columnHeight=[];
            var bodyWidth = document.body.offsetWidth;
            var n = parseInt(bodyWidth/inner.boxWidth);
            inner.columnNum = n==0?1:n;
            var bodyLeft = bodyWidth>=inner.boxWidth?bodyWidth-inner.columnNum*inner.boxWidth:0;
            inner.$con.css('left',parseInt(bodyLeft/2)-inner.marginOffset/2+'px');
            var i = 0;
            var l = inner.$boxes.length;
            for (i = 0; i <l ; i++) {
                if (i<inner.columnNum) {
                    inner.columnHeight[i]=inner.$boxes[i].offsetHeight+inner.marginOffset;
                    inner.$boxes.eq(i).css({
                        'top':0,
                        'left':i*inner.boxWidth+inner.marginOffset+'px'
                    });
                } else{
                    var innsertColumn = min(inner.columnHeight);
                    var imgHeight = inner.$boxes[i].offsetHeight+inner.marginOffset;
                    inner.$boxes.eq(i).css({
                        'top':inner.columnHeight[innsertColumn]+'px',
                        'left':innsertColumn*inner.boxWidth+inner.marginOffset+'px'
                    });
                    inner.columnHeight[innsertColumn] +=imgHeight;
                };
            };
            /* 重新定位之后需要重新计算父容器的高度，如果不明确赋值高度，会出现父容器高度为0，然后页面中的某些元素遮住瀑布流容器的情况 */
            if(inner.columnHeight.length>0){
                inner.$con.css('height',inner.columnHeight[max(inner.columnHeight)]+'px');
            }
            inner.index = i;
        };

        /* 初始化 */
        function init(){

            //由于瀑布流布局需要计算item的坐标然后定位，所以容器必须是relative，item必须是absolute
            inner.$con = $('#'+inner.id);
            inner.$con.css('position','relative');
            inner.boxWidth = inner.itemWidth+inner.marginOffset;

            bindEvents();
        }
        init();

        /* 绑定事件 */
        function bindEvents(){
            $(document).ready(function(){
                refreshAll();
                window.onresize = function(){
                    refreshAll();
                };
            });
        }

        /* 计算数组中的最小值的位置 */
        function min (heightArray) {
            var minColumn = 0;
            var minHeight = heightArray[minColumn];
            for (var i = 1,len=heightArray.length; i < len; i++) {
                var temp = heightArray[i];
                if (temp < minHeight) {
                    minColumn = i;
                    minHeight = temp;
                };
            };
            return minColumn;
        }
        /* 计算数组中的最大值的位置 */
        function max (heightArray){
            var maxColumn = 0;
            var maxHeight = heightArray[maxColumn];
            for(var i=0,len=heightArray.length;i<len;i++){
                var temp = heightArray[i];
                if(temp>maxHeight){
                    maxColumn = i;
                    maxHeight = temp;
                }
            }
            return maxColumn;
        }

        return component;
    };

    return waterfall;
}));
